#include <torch/extension.h>
#include <vector>
#include <cfloat>
#include <cuda_runtime.h>

// This kernel assumes that D, H, and W are exactly divisible by the corresponding pool sizes to avoid boundary checks.
// It uses fmaxf to compute the maximum value without divergent branching and precomputes the inverse of divisor to fuse division.
// Warp-level reduction with __shfl_down_sync is used to aggregate results uniformly and minimize warp divergence.

__global__ void fusedPoolingAndReductionKernel(
    const float* __restrict__ conv_out,  // (N, C, D, H, W)
    int N,
    int C,
    int D,
    int H,
    int W,
    int poolD,
    int poolH,
    int poolW,
    const float* __restrict__ bias,      // (C)
    int sum_dim,                         // if 1, reduce across channels
    float* __restrict__ output,          // if sum_dim==1, shape (N), else (N, C)
    float divisor
) {
    // Identify (n, c) slice by flattening blockIdx.x
    int nc = blockIdx.x;  // in [0, N*C)
    int n = nc / C;
    int c = nc % C;

    // Compute output sizes for pooling
    int outD = D / poolD;
    int outH = H / poolH;
    int outW = W / poolW;
    int numPools = outD * outH * outW;

    const float invDiv = 1.0f / divisor;  // Precompute inverse divisor
    float thread_sum = 0.0f;

    // Each thread processes multiple pooling windows in a strided loop with no divergent bounds
    for (int idx = threadIdx.x; idx < numPools; idx += blockDim.x) {
        int d = idx / (outH * outW);
        int rem = idx % (outH * outW);
        int h = rem / outW;
        int w = rem % outW;

        int startD = d * poolD;
        int startH = h * poolH;
        int startW = w * poolW;

        float max_val = -FLT_MAX;
        // Loop over pooling window; no boundary checks due to exact divisibility
        for (int pd = 0; pd < poolD; pd++) {
            for (int ph = 0; ph < poolH; ph++) {
                for (int pw = 0; pw < poolW; pw++) {
                    int curD = startD + pd;
                    int curH = startH + ph;
                    int curW = startW + pw;
                    int index = (((n * C + c) * D + curD) * H + curH) * W + curW;
                    // Fuse division via multiplication with invDiv
                    float val = conv_out[index] * invDiv;
                    max_val = fmaxf(max_val, val);
                }
            }
        }
        thread_sum += max_val;
    }

    // Intra-warp reduction without divergent branching
    unsigned int lane = threadIdx.x & 31;
    float sum_val = thread_sum;
    for (int offset = 16; offset > 0; offset /= 2) {
        sum_val += __shfl_down_sync(0xFFFFFFFF, sum_val, offset);
    }

    // Shared memory for inter-warp reduction
    extern __shared__ float shared[];
    if (lane == 0) {
        shared[threadIdx.x >> 5] = sum_val;
    }
    __syncthreads();

    float block_sum = 0.0f;
    int numWarps = (blockDim.x + 31) >> 5;
    if (threadIdx.x < numWarps) {
        block_sum = shared[threadIdx.x];
    }
    if (threadIdx.x < 32) {
        for (int offset = 16; offset > 0; offset /= 2) {
            block_sum += __shfl_down_sync(0xFFFFFFFF, block_sum, offset);
        }
    }

    // Final result: compute average and add bias in uniform control flow
    if (threadIdx.x == 0) {
        float avg = block_sum / numPools;
        avg = avg + bias[c];
        // Write either per (n,c) element or atomically reduce over channels
        if (sum_dim == 1) {
            atomicAdd(&output[n], avg);
        } else {
            output[n * C + c] = avg;
        }
    }
}

// Forward function: performs conv3d using cuDNN then the fused pooling and reduction

torch::Tensor forward_cuda(
    torch::Tensor x,
    double divisor,
    std::vector<int64_t> pool_size,  // pool_size: [poolD, poolH, poolW]
    int64_t sum_dim,
    torch::Tensor conv_weight,
    torch::Tensor conv_bias,
    torch::Tensor bias
) {
    TORCH_CHECK(x.is_cuda(), "x must be a CUDA tensor.");
    TORCH_CHECK(conv_weight.is_cuda(), "conv_weight must be a CUDA tensor.");
    TORCH_CHECK(conv_bias.is_cuda(), "conv_bias must be a CUDA tensor.");
    TORCH_CHECK(bias.is_cuda(), "bias must be a CUDA tensor.");

    // Use PyTorch's optimized conv3d (cuDNN) without fusion
    auto conv_out = at::conv3d(x, conv_weight, conv_bias);
    int N = conv_out.size(0);
    int C = conv_out.size(1);
    int D = conv_out.size(2);
    int H = conv_out.size(3);
    int W = conv_out.size(4);

    // Launch configuration: each block handles one (n, c) slice
    int totalBlocks = N * C;
    int threadsPerBlock = 128;
    int sharedMemSize = ((threadsPerBlock + 31) / 32) * sizeof(float);

    torch::Tensor output;
    if (sum_dim == 1) {
        // Sum over channel dimension, output shape: (N)
        output = torch::zeros({N}, conv_out.options());
    } else {
        // Output per (n, c) slice: shape (N, C)
        output = torch::empty({N, C}, conv_out.options());
    }

    fusedPoolingAndReductionKernel<<<totalBlocks, threadsPerBlock, sharedMemSize>>>(
        conv_out.data_ptr<float>(),
        N, C, D, H, W,
        pool_size[0], pool_size[1], pool_size[2],
        bias.data_ptr<float>(),
        sum_dim,
        output.data_ptr<float>(),
        static_cast<float>(divisor)
    );

    cudaDeviceSynchronize();
    return output;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward_cuda, "Fused 3D Conv, Pooling, GlobalAvgPool, BiasAdd, and Sum (CUDA) with uniform control flow");
}
